# Future
from __future__ import annotations


# isort: off
from spotipy.objects.base import *
# isort: on

# My stuff
from spotipy.objects.actions import *
from spotipy.objects.album import *
from spotipy.objects.artist import *
from spotipy.objects.category import *
from spotipy.objects.copyright import *
from spotipy.objects.device import *
from spotipy.objects.enums import *
from spotipy.objects.episode import *
from spotipy.objects.followers import *
from spotipy.objects.image import *
from spotipy.objects.playback import *
from spotipy.objects.playlist import *
from spotipy.objects.recommendation import *
from spotipy.objects.search import *
from spotipy.objects.show import *
from spotipy.objects.token import *
from spotipy.objects.track import *
from spotipy.objects.user import *
